#lang racket/base

(require redex/reduction-semantics)

(require "1-lc-syntax.rkt")

(provide (all-defined-out))

(define-extended-language lCKL lcL
  [v ::= ι (lambda (x) e)]
  [F ::= (apply hole e) (apply v hole) (let ([x hole]) e) (ifz hole e e)
     (op v ... hole e ...) ]
  [K ::= mt (F :: K)])

(define lck->
  (reduction-relation
   lCKL
   #:domain (e K)
   #:codomain (e K)

   (--> ((apply (lambda (x) e) v) K)
        ((substitute e x v) K)
        "β")

   (--> ((let ([x v]) e) K)
        ((substitute e x v) K)
        "ζ")

   (--> ((ifz 0 e_1 e_2) K)
        (e_1 K)
        "ifz")

   (--> ((ifz v e_1 e_2) K)
        (e_2 K)
        (side-condition (not (zero? (term v))))
        "ifnz")

   (--> ((op v ...) K)
        (v_1 K)
        (where v_1 (δ op v ...))
        "primop")

   (--> ((in-hole F e) K)
        (e (F :: K))
        (side-condition (not (redex-match? lCKL v (term e))))
        "push")

   (--> (v (F :: K))
        ((in-hole F v) K)
        "pop")))

(define (init-ck e)
  `(,e mt))

(module+ test
  (test-->>∃ lck->
            `(,running-example mt)
            (term
             ((apply f 1) ((+ 4 hole) :: mt)))))


;; ------------------------------------------------------------------------
;; ANF Machine

(require "2-anf-and-monadic-form.rkt")

(define anfck->
  (reduction-relation
   ANFL
   #:domain (M K)
   #:codomain (M K)

   (--> ((let ([x_1 (apply (lambda (x) M) V)]) M_1) K)
        ((substitute M x V) ((let ([x_1 hole]) M_1) :: K))
        "call")

   (--> (V (F :: K))
        ((in-hole F V) K)
        "return")

   (--> ((apply (lambda (x) M) V) K)
        ((substitute M x V) )
        "tail-call")

   (--> ((let ([x V]) M) K)
        ((substitute M x V) K)
        "move")

   (--> ((ifz 0 M_1 M_2) K)
        (M_1 K)
        "ifz")

   (--> ((ifz V M_1 M_2) K)
        (M_1 K)
        (side-condition (not (zero? (term V))))
        "ifnz")

   (--> ((let ([x (op V ...)]) M) K)
        ((substitute M x V_1) K)
        (where V_1 (δ op V ...))
        "primop")

   (--> ((op V ...) K)
        (V_1 K)
        (where V_1 (δ op V ...))
        "tail-primop")))

(require racket/match)
(define (stack-length K)
  (match K
    ['mt 0]
    [`(,F :: ,K)
     (add1 (stack-length K))]))

(define (ck-max-stack states)
  (apply max (map ck-stack-length states)))

(define (ck-stack-length state)
  (match state
    [`(,c ,K)
     (stack-length K)]))

(module+ test
  (require (only-in "2-anf-and-monadic-form.rkt" a-normalize))
  (define anf-eg1 (a-normalize running-example))
  (test-->>∃ anfck->
             `(,anf-eg1 mt)
             (term
              ((let ([x2 (apply f 1)]) (+ 4 x2)) mt)))

  (test-->>∃ anfck->
             (term
              ((let ((x1 (+ 2 2))) (let ((x 1)) (let ((x2 (+ x 1))) (+ x1 x2))))
               mt))
             '((+ 4 2) mt))

  (require rackunit)

  ;; CLAIM 1
  ;; ANF is an optimization; ANF uses strictly less stack space than \lambda in general.
  (check-equal?
   (ck-max-stack (apply-reduction-relation* #:all? #t anfck->
                                            `(,anf-eg1 mt)))
   0)

  (check-equal?
   (ck-max-stack (apply-reduction-relation* #:all? #t anfck->
                                            '((let ((x1 (+ 2 2))) (let ((x 1)) (let ((x2 (+ x 1))) (+ x1 x2))))
                                              mt)))
   0)

  (check-equal?
   (ck-max-stack (apply-reduction-relation* #:all? #t lck->
                                            `(,running-example mt)))
   2)

  (check-equal?
   (ck-max-stack (apply-reduction-relation* #:all? #t lck->
                                            `(,anf-eg1 mt)))
   2))


;; ------------------------------------------------------------------------
;; Region calculus

(define-extended-language regionL lcL
  [r ::= (variable-prefix r)]
  [e ::= .... (letregion r e) (@ r e)])

(define-union-language regionCK-baseL regionL lCKL)

(define-extended-language regionCSKL regionCK-baseL
  [F ::= .... (@ r hole)]
  [K ::= .... ((free r) :: K)]
  [o ::= (variable-prefix o)]
  [a ::= (r o)]
  [e ::= .... a]
  [v ::= .... a]
  [H ::= () (H (o v))]
  [S ::= () (S (r H))])

(module+ test
  (test-match regionCSKL e
              '(letregion r4
                          (apply
                           (let ([x (@ r1 2)])
                             (@ r4 (lambda (y) (@ r1 (* x y)))))
                           (@ r1 4)))))

(define-metafunction regionCSKL
  deref : S a -> v
  [(deref (S (r H)) (r o))
   (Hderef H o)]
  [(deref (S (r_!_! H)) ((name r r_!_1) o))
   (deref S (r o))])

(define-metafunction regionCSKL
  Hderef : H o -> v
  [(Hderef (H (o v)) o)
   v]
  [(Hderef (H (o_!_1 v)) (name o o_!_1))
   (Hderef H o)])

(define-metafunction regionCSKL
  alloc : S r v -> (S a)
  [(alloc S r v)
   ((insert S (r o) v) (r o))
   (where o ,(gensym 'o))])

(define-metafunction regionCSKL
  insert : S a v -> S
  [(insert (S (r H)) (r o) v)
   (S (r (H (o v))))]
  [(insert (S (r_1 H)) a v)
   ((insert S a v) (r_1 H))])

(define-metafunction regionCSKL
  free/ : S r -> S
  [(free/ (S (r H)) r)
   S]
  [(free/ (S ((name r r_!_1) H)) (name r_2 r_!_1))
   ((free/ S r_2) (r H))])

(define regioncsk->
  (reduction-relation
   regionCSKL
   #:domain (e S K)
   #:codomain (e S K)

   (--> ((apply a_1 a_2) S K)
        ((substitute e x a_2) S K)
        (where (lambda (x) e) (deref S a_1))
        "β")

   (--> ((let ([x a]) e) S K)
        ((substitute e x a) S K)
        "ζ")

   (--> ((ifz a e_1 e_2) S K)
        (e_1 S K)
        (where 0 (deref S a))
        "ifz")

   (--> ((ifz a e_1 e_2) S K)
        (e_2 S K)
        (where v (deref S a))
        (side-condition (not (zero? (term v))))
        "ifnz")

   (--> ((op a ...) S K)
        (v_1 S K)
        (where (v ...) ((deref S a) ...))
        (where v_1 (δ op v ...))
        "primop")

   (--> ((in-hole F e) S K)
        (e S (F :: K))
        (side-condition (not (redex-match? regionCSKL a (term e))))
        "push")

   (--> (a S (F :: K))
        ((in-hole F a) S K)
        "pop")

   (--> ((letregion r e) S K)
        (e (S (r ())) ((free r) :: K))
        "ralloc")

   (--> (v S ((@ r hole) :: K))
        (a S_1 K)
        (where (S_1 a) (alloc S r v))
        "alloc")

   (--> (a S ((free r) :: K))
        (a (free/ S r) K)
        "rfree")))

;; Read back the final address out of the machine state.
;; There must be exactly 1 final state.
(define (csk-read-val states)
  (match states
    [`((,a ,S mt))
     (term (deref ,S ,a))]
    [_ (error "terminated at invalid state" states)]))

(define (init-csk term)
  `(,term
    (() (r0 ()))
    mt))

(module+ test
  (define (final-state-equiv? state expected-val)
    (equal? (csk-read-val (list state)) expected-val))

  ;; r1 is the top-level return region
  (test-->> regioncsk->
            ;; should probably eliminate cycles
            #:equiv final-state-equiv?
            (init-csk '(letregion r4
                                  (apply
                                   (let ([x (@ r0 2)])
                                     (@ r4 (lambda (y) (@ r0 (* x y)))))
                                   (@ r0 4))))
            8))

(define-union-language anf-region-baseL regionL eval-lcL)
(define-extended-language aregionL anf-region-baseL
  [E ::= .... (@ r E)])

(define ra->
  ;; NOTE: Can't use extend-reduction-relation due to side conditions referring
  ;; directly to language names.
  ;; can't use where, because it doesn't support negation.
  (reduction-relation
   aregionL
   #:domain e
   #:codomain e
   (--> (in-hole E (let ([x e_1]) e_2))
        (let ([x e_1]) (in-hole E e_2))
        (side-condition
         (not (redex-match? aregionL hole (term E))))
        "A_1")
   (--> (in-hole E (ifz v e_1 e_2))
        (ifz v (in-hole E e_1) (in-hole E e_2))
        (side-condition (not (redex-match? aregionL hole (term E))))
        "A_2")
   (--> (in-hole E (op v ...))
        (let ([x (op v ...)])
          (in-hole E x))
        (side-condition
         (not (redex-match? aregionL hole (term E))))
        (side-condition
         (not (redex-match? aregionL (in-hole E_1 (let ([x_1 hole]) e)) (term E))))
        (side-condition
         (not (redex-match? aregionL (in-hole E_1 (@ r hole)) (term E))))
        (fresh x)
        "A_31")
   (--> (in-hole E (apply v ...))
        (let ([x (apply v ...)])
          (in-hole E x))
        (side-condition
         (not (redex-match? aregionL hole (term E))))
        (side-condition
         (not (redex-match? aregionL (in-hole E_1 (let ([x_1 hole]) e)) (term E))))
        (side-condition
         (not (redex-match? aregionL (in-hole E_1 (@ r hole)) (term E))))
        (fresh x)
        "A_32")
   ;; ------------------------------------------------------------------------
   ;; NEW RULES
   (--> (in-hole E (letregion r e))
        (letregion r (in-hole E e))
        (side-condition (not (redex-match? aregionL hole (term E))))
        "A_4")

   (--> (in-hole E (@ r v))
        (let ([x (@ r v)])
          (in-hole E x))
        (side-condition (not (redex-match? aregionL hole (term E))))
        (side-condition
         (not (redex-match? aregionL (in-hole E_1 (let ([x_1 hole]) e)) (term E))))
        (fresh x)
        "A_5")
   (--> (in-hole E (@ r (op v ...)))
        (let ([x (@ r (op v ...))])
          (in-hole E x))
        (side-condition (not (redex-match? aregionL hole (term E))))
        (side-condition
         (not (redex-match? aregionL (in-hole E_1 (let ([x_1 hole]) e)) (term E))))
        (fresh x)
        "A_51")
   (--> (in-hole E (@ r (apply v ...)))
        (let ([x (@ r (apply v ...))])
          (in-hole E x))
        (side-condition (not (redex-match? aregionL hole (term E))))
        (side-condition
         (not (redex-match? aregionL (in-hole E_1 (let ([x_1 hole]) e)) (term E))))
        (fresh x)
        "A_52")
   ))

(define ra->* (compatible-closure ra-> aregionL e))

(define reg1-term
  (term
   (letregion r2
              (@ r0 (* (letregion r1 (@ r2 (* (@ r1 1) (@ r1 2))))
                       (letregion r3 (@ r2 (* (@ r3 3) (@ r3 4)))))))))

(define anf-reg1-term
  (term
   (letregion
    r2
    (letregion
     r1
     (let ((x (@ r1 1)))
       (let ((x1 (@ r1 2)))
         (let ((x2 (@ r2 (* x x1))))
           (letregion
            r3
            (let ((x3 (@ r3 3)))
              (let ((x4 (@ r3 4)))
                (let ((x5 (@ r2 (* x3 x4))))
                  (@ r0 (* x2 x5)))))))))))))

(module+ test
  ;; this takes a long time (~10 seconds) to compute
  (test-match
   regionL
   (letregion
    r2
    (letregion
     r1
     (let ((x (@ r1 1)))
       (let ((x_1 (@ r1 2)))
         (let ((x_2 (@ r2 (* x x_1))))
           (letregion
            r3
            (let ((x_3 (@ r3 3)))
              (let ((x_4 (@ r3 4)))
                (let ((x_5 (@ r2 (* x_3 x_4))))
                  (@ r0 (* x_2 x_5)))))))))))

   ;; Returns a list of alpha-equivalent things.
   (car
    (apply-reduction-relation*
     ra->* #:cache-all? #t
     reg1-term))))

;; Return the length of a snoc list
(define (snoc-list-len ls)
  (let loop ([ls ls])
    (match ls
      [`()
       0]
      [`(,ls ,v)
       (add1 (loop ls))])))

(define (store-count-addrs S)
  (match S
    ['() 0]
    [`(,S (,r ,H))
     (+ (snoc-list-len H) (store-count-addrs S))]))

;; Count the maximum number of regions in the machine state
(define (csk-max-regions states)
  (apply max (map csk-regions-size states)))

(define (csk-regions-size state)
  (match state
    [`(,c ,S ,K)
     (snoc-list-len S)]))

;; Compute the maximum number of unfreed addresses in the machine state.
(define (csk-max-memory states)
  (apply max (map csk-memory-size states)))

(define (csk-memory-size state)
  (match state
    [`(,c ,S ,K)
     (store-count-addrs S)]))

(define reg1
  (init-csk reg1-term))

(define anf-reg1
  (init-csk anf-reg1-term))

(define-union-language m-region-baseL regionL meval-lcL)
(define-extended-language bregionL m-region-baseL
  [En ::= .... (@ r En)])

(define rb->
  (reduction-relation
   bregionL
   #:domain e
   #:codomain e
   (--> (in-hole En (let ([x e_1]) e_2))
        (let ([x e_1]) (in-hole En e_2))
        (side-condition
         (not (redex-match? bregionL hole (term En))))
        "B_1")
   (--> (in-hole En (ifz v e_1 e_2))
        (let ([x (ifz v e_1 e_2)])
          (in-hole En x))
        (side-condition
         (not (redex-match? bregionL hole (term En))))
        (fresh x)
        "B_2'")
   (--> (in-hole En (op v ...))
        (let ([x (op v ...)])
          (in-hole En x))
        (side-condition
         (not (redex-match? bregionL hole (term En))))
        (side-condition
         (not (redex-match? bregionL (in-hole En_1 (@ r hole)) (term En))))
        (fresh x)
        "B_31")
   (--> (in-hole En (apply v ...))
        (let ([x (apply v ...)])
          (in-hole En x))
        (side-condition
         (not (redex-match? bregionL hole (term En))))
        (side-condition
         (not (redex-match? bregionL (in-hole En_1 (@ r hole)) (term En))))
        (fresh x)
        "B_32")

   ;;------------------------------------------------------------------------
   ;; ------------------------------------------------------------------------
   ;; NEW RULES
   (--> (in-hole En (letregion r e))
        (let ([x (letregion r e)])
          (in-hole En x))
        (side-condition (not (redex-match? bregionL hole (term En))))
        (fresh x)
        "B_4")

   (--> (in-hole En (@ r v))
        (let ([x (@ r v)])
          (in-hole En x))
        (side-condition (not (redex-match? bregionL hole (term En))))
        (fresh x)
        "B_5")
   (--> (in-hole En (@ r (op v ...)))
        (let ([x (@ r (op v ...))])
          (in-hole En x))
        (side-condition (not (redex-match? bregionL hole (term En))))
        (fresh x)
        "B_51")
   (--> (in-hole En (@ r (apply v ...)))
        (let ([x (@ r (apply v ...))])
          (in-hole En x))
        (side-condition (not (redex-match? bregionL hole (term En))))
        (fresh x)
        "B_52")))

(define rb->* (compatible-closure rb-> bregionL e))

;; (apply-reduction-relation* rb->* #:cache-all? #t reg1-term)
;; this takes several seconds to run
(define bnf-reg1-term
  (term
   (letregion
    r2
    (let ((x4 (letregion
               r1
               (let ((x2 (@ r1 1)))
                 (let ((x3 (@ r1 2)))
                   (@ r2 (* x2 x3)))))))
      (let ((x5 (letregion
                 r3
                 (let ((x (@ r3 3)))
                   (let ((x1 (@ r3 4)))
                     (@ r2 (* x x1)))))))
        (@ r0 (* x4 x5)))))))

(define bnf-reg1
  (init-csk bnf-reg1-term))

(module+ test
  ;; monadic
  (test-match
   bregionL
   (letregion
    r2
    (let ((x_4 (letregion
               r1
               (let ((x_2 (@ r1 1)))
                 (let ((x_3 (@ r1 2)))
                   (@ r2 (* x_2 x_3)))))))
      (let ((x_5 (letregion
                 r3
                 (let ((x (@ r3 3)))
                   (let ((x_1 (@ r3 4)))
                     (@ r2 (* x x_1)))))))
        (@ r0 (* x_4 x_5)))))
   (car
    (apply-reduction-relation*
     rb->* #:cache-all? #t
     reg1-term)))

  ;; ANF correct
  (check-equal?
   (csk-read-val (apply-reduction-relation* regioncsk-> reg1))
   (csk-read-val (apply-reduction-relation* regioncsk-> anf-reg1)))

  ;; CLAIM: ANF extends the maximum number of active regions
  ;; i.e. extends the liftetime of some region
  (test-begin
    (define eg1-D (apply-reduction-relation* #:all? #t regioncsk-> reg1))
    (define anf-eg1-D (apply-reduction-relation* #:all? #t regioncsk-> anf-reg1))
    (check
     <
     (csk-max-regions eg1-D)
     (csk-max-regions anf-eg1-D)))

  ;; monadic correct
  (check-equal?
   (csk-read-val (apply-reduction-relation* regioncsk-> reg1))
   (csk-read-val (apply-reduction-relation* regioncsk-> bnf-reg1)))

  ;; CLAIM: monadic does not extends the maximum number of active regions
  (let ([eg1-D (apply-reduction-relation* #:all? #t regioncsk-> reg1)]
        [monadic-eg1-D (apply-reduction-relation* #:all? #t regioncsk-> bnf-reg1)])
    (check-equal?
     (csk-max-regions eg1-D)
     (csk-max-regions monadic-eg1-D))))

;; A non-trivial case involving lifting in predicate position.
(define-term lcr-region-in-pred-eg
  (letregion r1
             (ifz (letregion r2
                             (@ r1 (+ (@ r2 1) (@ r2 0))))
                  (@ r1 5)
                  (@ r1 6))))

(define mr-region-in-pred-eg
  (car (apply-reduction-relation* rb->*
                                  #:cache-all? #t
                                  (term lcr-region-in-pred-eg))))

(define anfr-region-in-pred-eg
  (car (apply-reduction-relation* ra->* #:cache-all? #t
                                  (term lcr-region-in-pred-eg))))

(module+ test
  (let ([lc-D (apply-reduction-relation* #:all? #t regioncsk->
                                         (init-csk (term lcr-region-in-pred-eg)))]
        [anf-D (apply-reduction-relation* #:all? #t regioncsk->
                                          (init-csk anfr-region-in-pred-eg))]
        [m-D (apply-reduction-relation* #:all? #t regioncsk->
                                        (init-csk mr-region-in-pred-eg))])
    (check-equal?
     (csk-max-regions lc-D)
     (csk-max-regions m-D))

    (check-equal?
     (csk-max-memory lc-D)
     (csk-max-memory m-D))

    (check
     <=
     (csk-max-regions m-D)
     (csk-max-regions anf-D))

    (check
     <
     (csk-max-memory m-D)
     (csk-max-memory anf-D))))

(module+ render-eg
  (require redex/gui)
  (traces regioncsk-> reg1))
